-- pure Lua server
-- (c) zöner

local socket   = require('socket')
local lfs      = require('lfs')

local server   = nil
local callback = nil
local clients  = {}
local time     = 0
local exit     = false
local tout     = 0
local static   = nil

-- request methods
local methods  = { 'OPTIONS', 'GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'TRACE', 'CONNECT' }

-- response codes
local codes = {
	[100]='Continue',               [101]='Switching Protocols', [200]='OK', [201]='Created', [202]='Accepted', [203]='Non-Authoritative Information',
	[204]='No Content',             [205]='Reset Content',       [206]='Partial Content', [300]='Multiple Choices', [301]='Moved Permanently',
	[302]='Found',                  [303]='See Other',           [304]='Not Modified', [305]='Use Proxy', [307]='Temporary Redirect', [400]='Bad Request',
	[401]='Unauthorized',           [402]='Payment Required',    [403]='Forbidden', [404]='Not Found', [405]='Method Not Allowed',
	[406]='Not Acceptable',         [407]='Proxy Authentication Required', [408]='Request Time-out', [409]='Conflict', [410]='Gone',
	[411]='Length Required',        [412]='Precondition Failed', [413]='Request Entity Too Large', [414]='Request-URI Too Large',
	[415]='Unsupported Media Type', [416]='Requested range not satisfiable', [417]='Expectation Failed',
	[500]='Internal Server Error',  [501]='Not Implemented', [502]='Bad Gateway', [503]='Service Unavailable',
	[504]='Gateway Time-out',       [505]='HTTP Version not supported',
}

-- content MIMEs
local mimes = {
	txt={name='text/plain; charset="utf-8"'}, bin={name='application/octet-stream', bin=true},
	html={name='text/html; charset="utf-8"'}, js={name='application/javascript; charset="utf-8"'},
	json={name='application/json; charset="utf-8"'}, jsonp={name='application/javascript; charset="utf-8"'},
	css={name='text/css; charset="utf-8"'},
	png={name='image/png',bin=true}, jpg={name='image/jpg',bin=true}, gif={name='image/gif',bin=true}, bmp={name='image/bmp',bin=true},
	tif={name='image/tiff',bin=true},
	ico={name='image/x-icon',bin=true}, pdf={name='application/pdf',bin=true},
}

-- mime aliases
local aliases = {
	txt='txt', text='txt', lst='txt', html='html', htm='html', css='css',
	js='js', javascript='js', jsonp='js', json='json',
	png='png', jpg='jpg', jpeg='jpg', gif='gif', bmp='bmp', tif='tif', tiff='tif',
	ico='ico', icon='ico',
	pdf='pdf',
}

-- add new client connection
local function addClient(_client)
	local i=1
	while clients[i] ~= nil do i=i+1 end
	clients[i] = {client=_client,
		req={ok=false, http={}, headers={}, body={}, cookies={}, post={}},
		resp={served=false, code=200, headers={}, cookies={}},
	}
end

-- split string by separator
local function split(s,sep)
	local t={}
	for ss in string.gmatch(s, '([^'..sep..']+)') do table.insert(t,ss) end
	return t
end
-- trim spaces
function trim(s)
	local n = s:find("%S")
	return n and s:match(".*%S", n) or ""
end

-- URL decode string
function urlDecode(s)
	s = string.gsub(s, '+', ' ')
	s = string.gsub(s, '%%(%x%x)', function(h) return string.char(tonumber(h,16)) end)
	s = string.gsub(s, '\r\n', '\n')
	return s
end
-- URL encode string
function urlEncode(str)
	if (str) then
		str = string.gsub(str, '\n', '\r\n')
		str = string.gsub(str, '([^%w %-%_%.%~])', function (c) return string.format ('%%%02X', string.byte(c)) end)
		str = string.gsub(str, ' ', "+")
	end
	return str
end

-- check if file exists
local function exists(fn)
--	local f = io.open(fn,'r')
--	if f ~= nil then io.close(f) return true else return false end
	local m,err = lfs.attributes(fn,'mode')
	return m == 'file'
end
-- get file extension
local function ext(fn)
	local ws = split(fn,'%.')
	if #ws > 1 then return string.lower(ws[#ws]) end
end

-- parse HTTP header
local function request(req)
	req.ok = false
--print('request lines:', #req.http)
	if #req.http < 2 then return end

	-- get method,path,protocol
	local ws = split(req.http[1],'%s')
	if #ws ~= 3 then return end
	req.method    = ws[1]
	req.path      = ws[2]
	req.protocol  = ws[3]

	-- get fragment,query
	ws = split(req.path,'%#')
	req.fragment = ws[2]
--print('  hash:',req.hash)
	ws = split(ws[1],'%?')
	req.query   = ws[2]
	req.path    = ws[1]
	-- URL extension
	req.ext     = aliases[ext(req.path)]

	req.queries = {}
	if req.query ~= nil then for k,v in pairs(split(req.query,'%&')) do
		ws = split(v,'%=')
		if #ws == 2 then req.queries[ws[1]] = trim(ws[2]) end
	end end

	-- get headers and body
	for i=2,#req.http do
		if string.len(req.http[i]) ~= 0 then
			ws = split(req.http[i],'%s')
			if (#ws == 2) and (string.len(ws[1]) > 0) and (string.sub(ws[1],-1,-1) == ':') then
				req.headers[string.sub(ws[1],1,-2)] = trim(ws[2])
			end
		end
	end

	-- get cookies
	if req.headers['Cookie'] ~= nil then
		for k,v in pairs(split(req.headers['Cookie'],'%;')) do
			ws = split(v,'%=')
			if #ws == 2 then req.cookies[ws[1]] = trim(ws[2]) end
		end
	end

	-- get POST parameters
	if (req.method == 'POST') and (#req.body > 0) then
		for k,v in pairs(split(req.body[1],'%&')) do
			ws = split(v,'%=')
			if #ws == 2 then req.post[ws[1]] = trim(ws[2]) end
		end
	end

	req.ok = true
end

-- generate response
local function respond(req, resp, callback, static)
	if not req.ok then return end

	-- set MIME
	if req.ext ~= nil then resp.headers['Content-Type'] = mimes[req.ext].name end

	-- serve file
	local fn = static .. req.path
print('request url:',fn)
	if (req.method == 'GET') and (static ~= nil) and (exists(fn)) then
		local f = io.open(fn,'r')
		if (req.ext ~= nil) and (f ~= nil) then
print('serving file:',fn,req.ext)
			if mimes[req.ext].bin == true then resp.blob = f:read('*a')
			else
				resp.text = {}
				local l = f:read('*l')
				while l ~= nil do
					table.insert(resp.text,l)
					l = f:read('*l')
				end
			end
			io.close(f)
			resp.code = 200
			resp.served = true
		end
	end

--print('served:',resp.served,'callback:',callback)
	-- serve dynamic content
	if (not resp.served) and (callback ~= nil) then
		callback(req,resp)
	end

	-- none served - return error 404
	if not resp.served then resp.code = 404	end

	-- generate HTTP response
	--   1st line
	local data = req.protocol .. ' ' .. resp.code .. ' ' .. codes[resp.code] .. '\r\n'
	if resp.blob ~= nil then resp.headers['Content-Length'] = string.len(resp.blob) end
	--   headers
	for k,v in pairs(resp.headers) do
		data = data .. k .. ': ' .. v .. '\r\n'
	end
	--   cookies
	for k,v in pairs(resp.cookies) do
		data = data .. 'Set-Cookie: ' .. k .. '=' .. v .. '\r\n'
	end
	--   extra empty line
	data = data .. '\r\n'
	--   append text body
	if resp.text ~= nil then
		if type(resp.text) == 'string' then data = data .. resp.text .. '\r\n'
		else
			for k,v in pairs(resp.text) do
				data = data .. v .. '\r\n'
			end
		end
		data = data .. '\r\n'
	end
	--   append binary body
	if resp.blob ~= nil then
		data = data .. resp.blob
	end
	resp.data = data
	resp.served = true
end

-- dispatcher for new/existing connections
local function dispatch()
	while not exit do

	-- server closed - exit
	if server == nil then return end

	-- add new connection
	local _client = server:accept()
	if _client ~= nil then
		_client:settimeout(0, 'b')
		_client:settimeout(0, 't')
		addClient(_client)
		print('new connection')
	end

	-- dispatch opened connections
	for k,cl in pairs(clients) do
		local data,err = cl.client:receive('*l')
		if data ~= nil then
			table.insert(cl.req.http, data)

			-- last line? parse HTTP packet
			if (string.len(data) == 0) and (#cl.req.http > 2) then

				-- receive rest of body
				while true do
					local data,err = cl.client:receive('*l')
					if data ~= nil then	table.insert(cl.req.body, data)
					else break end
				end

if #cl.req.body > 0 then
	print('\r\nbody:')
	for k,v in pairs(cl.req.body) do
		print('  ' .. v)
	end
end

				request( cl.req )
				respond( cl.req, cl.resp, callback, static )
				if cl.resp.served then -- send response
--print('serving content:\r\n\r\n' .. clients[i].resp.data)
					cl.client:send(cl.resp.data)
					cl.client:close()
				end
			end
		else
			if err == 'closed' then
				cl = nil -- delete closed connection
			else
				-- print('error:',err)
			end
		end
		coroutine.yield(true)
	end

	coroutine.yield(true)
	end

	-- close opened connections
	for k,cl in pairs(clients) do cl.client:close() end
end

local function serve(_port, _static, _conn, _callback)
	if _static ~= nil then static = _static end
	if _conn == nil then _conn = 64 end
	callback = _callback
	if server ~= nil then server:close() end
	server = socket.tcp()
	server:bind('*', _port)
	server:listen(_conn)
	server:settimeout(0, 'b')
	server:settimeout(0, 't')

	return coroutine.wrap(dispatch)
end

-- close server connection
local function close()
	exit = true
end

-- return module structure
return {
	methods = methods,
	codes   = codes,
	mimes   = mimes,

	serve   = serve,
	close   = close,
}

